/* $Id: acCmd.cpp for Odin
 * $auth: Steve Torchinsky
 * $created: Calgary, Tuesday, 2001 May 15, 16:16:08
 *
 * This is the part of the cpp version of mkbckseq
 * mkbckseq is a Bash script employing sed and gawk which
 * uses templates to create backend command scripts
 *
 * The cpp version will do the same job, and will also read
 * the mode configurations from text files exported from the
 * Radiometer_modes.xls excel spreadsheet by Urban Frisk
 *
 * 
 * definition of the methods for class acCmd
 *
 * changes:
 *   Calgary, Monday, 2001 November 12, 06:31:08
 *   new default mode for AC1 8-band to put dead sub-band at edge
 *
 *   Calgary, Wednesday, 2001 December 05, 16:31:30
 *   trying to implement new spreadsheet version which includes
 *   correlator mux and ssb settings in the obsmode sheet
 *
 *   Calgary, Thursday, 2001 December 06, 12:07:27
 *   change to method setMode:  argument is mux instead of nbands
 *
 *   Calgary, Thursday, 2001 December 06, 14:04:08
 *   change to method nbands() to count bands from mux setting
 * 
 *   Calgary, Tuesday, 2002 February 05, 13:50:15
 *   v2.4.4
 *   adding methods for attenuator settings
 *
 *   Calgary, Friday, 2002 February 15, 14:53:54
 *   removing all references to modeIndx, N_ACMODES, etc
 *
 *   Calgary, Wednesday, 2002 February 20, 10:19:09
 *   adding attenuator settings to default settings
 *   for AC1, the bad band is SSB1, attenuation is set to max
 *   for AC2, the bad band is SSB2, attenuation is set to max
 */

#include <cstdlib>
#include <strstream>
#include <iomanip>
#include "OdinCmd.h"
#include "acCmd.h"

/* constructor */
acCmd::acCmd(unsigned short _whichAC){
  switch (_whichAC){
  case AC1:
    whichAC=AC1;
    usr=AC1USR;
    ACn=1;
    break;
  case AC2:
    whichAC=AC2;
    usr=AC2USR;
    ACn=2;
    break;
  default:
    whichAC=AC1;
    usr=AC1USR;
    ACn=1;
  }
  if (!defaultSettings())
    std::cout << "\nacCmd::acCmd  WARNING could not finish default settings";
}

/* destructor */
acCmd::~acCmd(){
}

/* method to assign correlator mode
 * AC1 ssb frequencies are high,low,high,low 
 * AC2 ssb frequencies are low,high,low,high 
 * ssb oscillator ranges are 3400 - 3900 and 3800 - 4400
 * do a check to make sure we are using acceptable values
 */
bool acCmd::setMode(unsigned short _mode, unsigned short _mux,
		    unsigned short ssb1, unsigned short ssb2, 
		    unsigned short ssb3, unsigned short ssb4, 
		    char *modename){

  switch(whichAC){
  case AC1:
    if ((ssb1 < 3800) || (ssb1 > 4400))return setModeError(AC1,1,ssb1);
    if ((ssb2 < 3400) || (ssb2 > 3900))return setModeError(AC1,2,ssb2);
    if ((ssb3 < 3800) || (ssb3 > 4400))return setModeError(AC1,3,ssb3);
    if ((ssb4 < 3400) || (ssb4 > 3900))return setModeError(AC1,4,ssb4);
    break;

  case AC2:
    if ((ssb1 < 3400) || (ssb1 > 3900))return setModeError(AC2,1,ssb1);
    if ((ssb2 < 3800) || (ssb2 > 4400))return setModeError(AC2,2,ssb2);
    if ((ssb3 < 3400) || (ssb3 > 3900))return setModeError(AC2,3,ssb3);
    if ((ssb4 < 3800) || (ssb4 > 4400))return setModeError(AC2,4,ssb4);
    break;

  default:
    cout << "\nERROR: trying to specify non-existent correlator\n";
    return false;

  }

  /* assign the mode number */
  mode=_mode;

  /* assign the mux command. 
   * It must be shifted 8 bits to be correct in the command data
   */
  mux=(_mux << 8);

  /* assign the mode description */
  modeDescription=modename;

  /* assign the SSB frequencies */
  ssbFreq[0]=ssb1;
  ssbFreq[1]=ssb2;
  ssbFreq[2]=ssb3;
  ssbFreq[3]=ssb4;


  return true;

}

/* a method to return an error message when setMode fails
 * because of unacceptable SSB settings
 */
bool acCmd::setModeError(unsigned short whichAC,
			 unsigned short whichSSB,
			 unsigned short ssb){
  switch (whichAC){
  case AC1:
    cout << "\nERROR:  inappropriate SSB settings for AC1:  ";
    break;

  case AC2:
    cout << "\nERROR:  inappropriate SSB settings for AC2:  ";
    break;

  default:
    cout << "\nERROR:  unknown correlator set mode failure:  ";
  }

  cout << "SSB" << whichSSB << "=" << ssb << " is out of range.";
  return false;
}

bool acCmd::defaultSettings(){
  setIntTime(0);
  clearMasks();
  enableSwitching();
  selectInputChannel(CH119);
  setPrescaler(2);
  setNref(0);
  setNsig(0);
  selectMode(1);

  switch(whichAC){
  case AC1:
    return defaultSettingsAC1();
    break;

  case AC2:
    return defaultSettingsAC2();
    break;

  default:
    return false;
  }

}

/* default settings for modes up to 13 taken from 
 * Radiometer_modes_v28.2.xls sent to me by Gunnar 2001Dec05
 *
 * Calgary, Friday, 2002 February 15, 14:57:10 
 *  changing defaultSettingsAC? to be a switch statement instead of filling
 *  an array with all the default modes
 * 
 *  mode number should already be assigned before calling defaultSettings...
 *
 * 
 */

/* default settings for AC1: ssb frequencies are high,low,high,low */
bool acCmd::defaultSettingsAC1(){
  disableFilterbank();

  /* attenuator settings:  SSB1 is the bad band for AC1 */
  att[0]=AC_MAX_ATT;
  att[1]=AC_NOM_ATT;
  att[2]=AC_NOM_ATT;
  att[3]=AC_NOM_ATT;

  switch (mode) {
  case 1:
    if (MKBCKSEQ_VERSION_MINOR == 3)
    /* new 8-band mode because of dead sub-band 2001Nov12 *
     * see email from Michael Olberg today                *
     * LO1 = 4250 MHz, LO3 = 4050 MHz, LO2 = 3650 MHz, LO4 = 3850 MHz
     * 
     * this is not yet implemented on the satellite as of 21Nov2001
     *
     * Calgary, Thursday, 2001 December 06, 16:35:40
     * this is now superseded by correlator modes being 
     * defined in the spreadsheet
     */
      return setMode( 1,AC_MODE_8,4250,3650,4050,3850,"1 x 800 MHz");
    else
      return setMode( 1,AC_MODE_8,4000,3600,4200,3800,"1 x 800 MHz");
   break;
    
  case 2:
    return setMode( 2,AC_MODE_8,4100,3500,4300,3700,"2 x 400 BW 1.0GHz");
    break;

  case 3:
    return setMode( 3,AC_MODE_8,4200,3400,4400,3600,"2 x 400 BW 1.2GHz");
    break;

  case 4:
    return setMode( 4,AC_MODE_4,4000,3800,4000,3800,"1 x 400 MHz");
    break;

  case 5:
    return setMode( 5,AC_MODE_2,3900,3400,3900,3400,"1 x 200 MHz");
    break;

  case 6:
    return setMode( 6,AC_MODE_1,3849,3400,4400,3400,"1 x 100 MHz");
    break;

  case 7:
    return setMode( 7,AC_MODE_8,4200,3700,4400,3900,"800 MHz shift high");
    break;

  case 8:
    return setMode( 8,AC_MODE_4,4100,3500,4300,3700,"2 x 200 BW 1.0GHz");
    break;

  case 9:
    return setMode( 9,AC_MODE_8,4300,3700,4100,3900,"s1a: 700MHz");
    break;

  case 10:
    return setMode(10,AC_MODE_8,4200,3600,4000,3800,"sm1a: 700MHz");
    break;
      
  case 11:
    return setMode(11,AC_MODE_8,4000,3600,4200,3800,"w3a,w4a: 500, 200 MHz");
    break;

  default:
    return false;
  }
  return false;
}


/* default settings for AC2: ssb frequencies are low,high,low,high */
bool acCmd::defaultSettingsAC2(){
  enableFilterbank();

  /* attenuator settings:  SSB2 is the bad band for AC2 */
  att[0]=AC_NOM_ATT;
  att[1]=AC_MAX_ATT;
  att[2]=AC_NOM_ATT;
  att[3]=AC_NOM_ATT;

  switch (mode) {
  case 1:
    return setMode( 1,AC_MODE_8,3600,4000,3800,4200,"1 x 800 MHz");
    break;

  case 2:
    return setMode( 2,AC_MODE_8,3500,4100,3700,4300,"2 x 400 BW 1.0GHz");
    break;

  case 3:
    return setMode( 3,AC_MODE_8,3400,4200,3600,4400,"2 x 400 BW 1.2GHz");
    break;

  case 4:
    return setMode( 4,AC_MODE_4,3800,4000,3800,4000,"1 x 400 MHz"); 
    break;

  case 5:
    return setMode( 5,AC_MODE_2,3900,4400,3900,4400,"1 x 200 MHz");
    break;

  case 6:
    return setMode( 6,AC_MODE_1,3849,4400,3400,4400,"1 x 100 MHz");
    break;

  case 7:
    return setMode( 7,AC_MODE_8,3700,4200,3900,4400,"800 MHz shift high");
    break;

  case 8:
    return setMode( 8,AC_MODE_4,3500,4100,3700,4300,"2 x 200 BW 1.0GHz");
    break;

  case 9:
    return setMode( 9,AC_MODE_4,3600,4200,3800,4000,"sm1a: 700 MHz");
    break;

  case 10:
    return setMode(10,AC_MODE_4,3900,4400,3400,4400,"ast,119:  100 MHz+50MHz");
    break;

  case 11:
    return setMode(11,AC_MODE_4,3400,4200,3600,4400,"s1a: 400,100,200 MHz");
    break;

  case 12:
    return setMode(12,AC_MODE_4,3500,4100,3700,4300,"w3a: 400,100,300 MHz");
    break;

  case 13:
    return setMode(13,AC_MODE_4,3600,4000,3800,4200,"w4a: 500,200 MHz");
    break;

  default:
    return false;
  }
  return false;
}

/* private method, return the number of bands for this mode
 *
 * the mux setting gives the subbands as follows:
 * bit8: SSB1 usb 
 * bit7: SSB1 lsb 
 * bit6: SSB2 usb 
 * bit5: SSB2 lsb 
 * bit4: SSB3 usb 
 * bit3: SSB3 lsb 
 * bit2: SSB4 usb 
 * bit1: SSB4 lsb
 *
 *  BUT, bit8 is always equal to zero (even though SSB1-usb is always on)
 *  for example:  01111111=0x7f is 8-band mode
 *                00101010=0x2a is 4-band mode (all SSB's upper side band)
 */
unsigned short acCmd::nbands(){
  short bandCounter=1;
  unsigned short muxVal= (mux>>8);
  for (short i=0;i<8;i++){
    if( (muxVal & 1)==1 )bandCounter++;
    muxVal=(muxVal>>1);
  }
  return bandCounter;
}

/* public method, make the mode selection */
bool acCmd::selectMode(unsigned short _mode){
  mode = _mode;
  return true;
}

/* public method, select the input channel */
bool acCmd::selectInputChannel(unsigned short _channel){
  channel = _channel;
  switch (channel){
  case CH495: iport=2;
    break;
  case CH549: iport=1;
    break;
  case CH555: iport=4;
    break;
  case CH572: iport=3;
    break;
  case CH119: iport=6;
    break;
  case CHCOMBINER: iport=5;
    break;
  default:
    return false;
  }
  return true;
}

/* public method, select input port
 * (same as above, but choose port instead of frontend channel)
 */
bool acCmd::selectInputPort(unsigned short _iport){
  iport=_iport;
  switch (iport){
  case 2: channel=CH495;
    break;
  case 1: channel=CH549;
    break;
  case 4: channel=CH555;
    break;
  case 3: channel=CH572;
    break;
  case 6: channel=CH119;
    break;
  case 5: channel=CHCOMBINER;
    break;
  default:
    return false;
  }
  return true;
}

/* public method, return the currently selected input channel */
unsigned short acCmd::getInputChannel(){
  return channel;
}

/* public method, return the currently selected input port */
unsigned short acCmd::getInputPort(){
  return iport;
}

/* public method, enable switching */
bool acCmd::enableSwitching(){
  switching=AC_SWITCH_ON;
  return true;
}

/* public method, disable switching */
bool acCmd::disableSwitching(){
  switching=AC_SWITCH_OFF;
  return true;
}

/* public method, enable filterbank */
bool acCmd::enableFilterbank(){
  if (whichAC==AC2)filterbank=AC_FILTER_ON;
  else filterbank=AC_FILTER_OFF;
  return true;
}

/* public method, disable filterbank */
bool acCmd::disableFilterbank(){
  filterbank=AC_FILTER_OFF;
  return true;
}

/* public method, set integration time */
bool acCmd::setIntTime(unsigned short _intTime){
  intTime=_intTime;
  return true;
}

/* private method, clear all sync masks */
bool acCmd::clearMasks(){
  pointingMask=0;
  chopMask=0;
  F119Mask=0;
  return true;
}

/* public method, set mask */
bool acCmd::setMask(unsigned short whichMask, unsigned short mask){
  if ((whichMask & MASK_ACDC1) == MASK_ACDC1)
    pointingMask=pointingMask | mask;

  if ((whichMask & MASK_ACDC2) == MASK_ACDC2)
    pointingMask=pointingMask | (mask << 8);

  if ((whichMask & MASK_CHOPA) == MASK_CHOPA)
    chopMask=chopMask | mask;

  if ((whichMask & MASK_CHOPB) == MASK_CHOPB)
    chopMask=chopMask | (mask << 8);

  if ((whichMask & MASK_119) == MASK_119)
    F119Mask=F119Mask | mask;

  return true;
}

/* public method, set prescaler */
bool acCmd::setPrescaler(unsigned short _prescaler){
  if ( (_prescaler<2) || (_prescaler>6) )
    prescaler=2;
  else
    prescaler=_prescaler;
  return true;
}

/* public method, set number of on-board sig integrations */
bool acCmd::setNsig(unsigned short _nsig){
  nsig=_nsig;
  return true;
}

/* public method, set number of on-board ref integrations */
bool acCmd::setNref(unsigned short _nref){
  nref=_nref;
  return true;
}

/* public method, set the attenuator setting */
bool acCmd::setAtt(short n, unsigned short val){
  if ((n<0) || (n>3))return false;
  att[n]=val;
  return true;
}

/* METHODS FOR GENERATING COMMAND STRINGS FOR OCOM */

/* private method, make a correlator command */
char *acCmd::cmd(unsigned short addr, unsigned short data, char *comment){
  OdinCmd ocom;
  strstream mkComment;
  mkComment << "AC"
	    << dec << setw(1) 
	    << ACn
	    << ":"
	    << comment
	    << ends;
  return  ocom.cmd(usr,addr,data,mkComment.str());
}

/* public method, make config command */
char *acCmd::configCmd(){
  return cmd(AC_CONFIG,AC_CONFIG_DAT,"Config");
}

/* public method, make boot command */
char *acCmd::bootCmd(){
  return cmd(AC_BOOT,AC_BOOT_DAT,"Boot");
}

/* public method, make halt command */
char *acCmd::haltCmd(){
  return cmd(AC_HALT,AC_HALT_DAT,"Halt");
}

/* public method, make start command */
char *acCmd::startCmd(){
  return cmd(AC_START,AC_START_DAT,"Start");
}

/* public method, make calibrate command */
char *acCmd::calibrateCmd(){
  return cmd(AC_CALIBRATE,AC_CALIBRATE_DAT,"Calibrate");
}

/* public method, make the SSB commands */
char *acCmd::ssbCmd(){
  strstream *mkCommand = new strstream;
  for (unsigned short i=0;i<4;i++){
    strstream *mkComment = new strstream;
    *mkComment << "SSB" << dec << setw(1) << i+1 
	       << " Freq: " << dec << setw(4) << ssbFreq[i]
	       << " MHz"
	       << ends;
    *mkCommand << cmd(AC_SSB1-i,ssbFreq[i],mkComment->str());
    delete mkComment;
  }

  *mkCommand << ends;
  return mkCommand->str();
}

/* public method, make attenuator command */
char *acCmd::attCmd(short n){
  if ((n<0)||(n>3))
    return "! ERROR\n! ERROR:  INVALID ATTENUATOR SETTING\n! ERROR";

  strstream mkComment;
  mkComment << "ATT" << dec << setw(1) << n+1 << " "
	    << dec << setw(4) << att[n]
	    << ends;
  return cmd(AC_ATT1+n,att[n],mkComment.str());
}

/* public method, make ALL attenuator commands */
char *acCmd::attenuatorsCmd(){
  strstream mkCmd;
  for (short n=0;n<4;n++)
    mkCmd << attCmd(n);
  mkCmd << ends;
  return mkCmd.str();
}

/* public method, make input/switching/filterbank command */
char *acCmd::inputCmd(){
  strstream *mkComment=new strstream;
  unsigned short data=switching | (iport << 8) | filterbank;
  switch (switching){
  case AC_SWITCH_ON:
    *mkComment << "switching";
    break;
  case AC_SWITCH_OFF:
    *mkComment << "no switching";
    break;
  default:
    *mkComment << "INVALID SWITCHING";
  }
  *mkComment << ";Input "
	     << dec << setw(1) << iport
	     << " (";

  if (channel==CHCOMBINER)
    *mkComment << "Sig combiner)";
  else
    *mkComment << dec << setw(3) 
	       << channel
	       << ")";

  /* filter bank should only be enabled for AC2, 
   * but it's checked already in method enableFilterbank
   */
  if (whichAC==AC2){
    if (filterbank==AC_FILTER_ON)
      *mkComment << ";filter bank on";
    else
      *mkComment << ";filter bank off";
  }

  *mkComment << ends;

  return cmd(AC_INPUT,data,mkComment->str());
}

/* public method, make input/switching/filterbank command
 * this simply calls the inputCmd method, but it's defined here
 * with filter and switching names,  for convenience 
 */
char *acCmd::switchingCmd(){
  return inputCmd();
}

char *acCmd::filterCmd(){
  return inputCmd();
}

/* public method, make mux/integration time command */
char *acCmd::modeCmd(){
  unsigned short data=(intTime << 4) | mux;
  strstream *mkComment=new strstream;
  *mkComment << "Corr chip mux set: "
	     << dec << setw(1) << nbands()
	     << " bands; Int time:  "
	     << dec << setw(2) << intTime
	     << " secs"
	     << ends;
  return cmd(AC_MODE,data,mkComment->str());
}

/* public method, mux/integration time command, with int time name */
char *acCmd::intTimeCmd(){
  return modeCmd();
}

/* public method, construct synch mask command */
char *acCmd::maskCmd(unsigned short whichMask){
  strstream *mkCommand=new strstream;

  if ( ((whichMask & MASK_ACDC1) == MASK_ACDC1)
       || ((whichMask & MASK_ACDC2) == MASK_ACDC2))
    *mkCommand << cmd(AC_MASKACDC1,pointingMask,
		      "Pointing:Hi byte ACDC2, Low byte ACDC1");

  if ( ((whichMask & MASK_CHOPA) == MASK_CHOPA)
       || ((whichMask & MASK_CHOPB) == MASK_CHOPB))
    *mkCommand << cmd(AC_MASKCHOPA,chopMask,
		      "Chop Sync:Hi byte MechB, Low byte MechA");

  if ((whichMask & MASK_119) == MASK_119)
    *mkCommand << cmd(AC_MASK119,F119Mask,
		      "119 frequency switch Mask");

  *mkCommand << ends;
  return mkCommand->str();

}

/* public method, construct command to clear all masks */
char *acCmd::clearMasksCmd(){
  clearMasks();
  return maskCmd(MASK_ACDC1 + MASK_ACDC2 + MASK_CHOPA + MASK_CHOPB + MASK_119);
}

/* public method, construct correlator identity command */
char *acCmd::idCmd(){
  strstream *mkComment = new strstream;
  *mkComment << "User ID: "
	     << hex << setw(1) << setiosflags(ios::uppercase)
	     << usr
	     << ends;
  return cmd(ACID,usr,mkComment->str());
}

/* public method, construct prescaler command */
char *acCmd::prescalerCmd(){
  strstream *mkComment = new strstream;
  *mkComment << "Prescaler: "
	     << dec << setw(1) << prescaler
	     << " (valid values are 2 to 6)"
	     << ends;
  return cmd(AC_PRESCALE,prescaler,mkComment->str());
}

/* public method, construct sig/ref command */
char *acCmd::sigRefCmd(){
  strstream *mkComment = new strstream;
  *mkComment << "Sig spectras: S="
	     << dec << setw(2) << nsig
	     << ":Ref spectras: R="
	     << dec << setw(2) << nref
	     << ends;
  return cmd(AC_SIGREF,(nsig << 8) | nref,mkComment->str());
}

/* public method, produce correlator configuration commands
 * setting up the correlator, and preparing for a Gain Search
 */
char *acCmd::configurationCmd(){
  strstream *mkSeq=new strstream;
  disableSwitching();

  *mkSeq 
    << "\n"
    << "! AC"
    << dec << setw(1) << ACn
    << " Identification\n"
    << idCmd()
    << "\n! Clear Synch masks for gain search\n"
    << clearMasksCmd()

    << "\n!AC"
    << dec << setw(1) << ACn
    <<" MODE\n"
    << modeCmd()
    << inputCmd()
    << sigRefCmd()
    << prescalerCmd()

    << "\n!AC"
    << dec << setw(1) << ACn
    <<" SSB SETTINGS\n"
    << ssbCmd()
    << "\n\n! UPLOAD CONFIGURATION TO CORRELATOR\n"
    << configCmd()
    << "\n"
    << ends;

  return mkSeq->str();
}



