/* $Id: aosCmd.cpp for Odin
 * $auth: Steve Torchinsky
 * $created: Calgary-to-Toronto, Tuesday, 2001 May 08, 12:03:53
 *
 * This is part of the cpp version of mkbckseq
 * mkbckseq is a Bash script employing sed and gawk which
 * uses templates to create backend command scripts
 *
 * The cpp version will do the same job, and will also read
 * the mode configurations from text files exported from the
 * Radiometer_modes.xls excel spreadsheet by Urban Frisk
 *
 * 
 * the methods for class aosCmd are defined here
 *
 * changes:
 *  Calgary, Wednesday, 2001 September 05, 10:28:24
 *  fixed bug with AOS input:  
 *    spreadsheet gives port number which must be translated to 2^port-1
 *    for the AOS command
 *
 *  Calgary, Monday, 2001 September 24, 13:32:46
 *    default temperature control level is changed from 19.12C to 21C
 *
 *  Calgary, Friday, 2001 November 09, 08:46:10
 *    default temperature control level is changed to 22C
 *    (email from ALX today)
 *
 *  Calgary, Friday, 2001 November 09, 11:25:48
 *    temperature control level is rounded up to nearest integer
 *    instead of rounding down
 *
 *  Calgary, Friday, 2001 November 09, 11:28:33
 *    BAD BUG:  temperature control level command was using attenuator address!
 *
 *  Calgary, Thursday, 2001 November 22, 12:07:46
 *    temperature control command removed altogether.
 *    it will be set by the power-on script
 *    see email from Stefan Lundin today
 *
 * Calgary, Friday, 2001 December 07, 08:46:43
 *   temperature control command re-inserted
 *   they're not yet ready to eliminate it
 *   see email from Gunnar today
 *                              
 * Calgary, Wednesday, 2002 January 16, 07:53:01
 *   temperature control command removed (commented out)
 *   see email from Gunnar today
 *
 * Calgary, Wednesday, 2002 February 20, 11:04:44
 *   using macro definition for max attenuator setting:  AOS_MAX_ATT
 *
 *  Calgary, Wednesday, 2002 February 20, 16:02:26
 *   added definition for method setAttenuator()
 *
 *  Calgary, Wednesday, 2002 February 20, 17:05:24
 *   new private method attCmd()
 *   method attenuatorCmd() is overloaded
 */

#include <cstdlib>
#include <fstream>
#include <strstream>
#include <iomanip>
#include <cstring>
#include <cmath>
#include "aosCmd.h"

/* constructor */
aosCmd::aosCmd(){
  defaultSettings();
}

/* destructor */
aosCmd::~aosCmd(){
}

/* private method, default settings */
bool aosCmd::defaultSettings(){
  channel=CH119;
  intTime=0;
  laserCurrent=86.24;
  tempControl=22.0;
  attenuator=AOS_NOM_ATT;
  return true;
}

/* public method, select output mode */
bool aosCmd::selectOutputMode(unsigned short _outputMode){
  if (_outputMode>7) outputMode=AOS_MODE_32;
  else outputMode=_outputMode;
  return true;
}

/* private method, translate input port to channel */
unsigned short aosCmd::channel2iport(){
  switch (channel) {
  case CH495: return AOS_495;
  case CH549: return AOS_549;
  case CH555: return AOS_555;
  case CH572: return AOS_572;
  case CH119: return AOS_119;
  default: return AOS_NOINPUT;
  }
  return 0;
}

/* private method, translate input channel to port number */
unsigned short aosCmd::iport2channel(unsigned short _port){
  switch (_port) {
  case AOS_495: return CH495;
  case AOS_549: return CH549;
  case AOS_555: return CH555;
  case AOS_572: return CH572;
  case AOS_119: return CH119;
  default: return NOINPUT;
  }
  return 0;
}

/* public method, assign input channel */
bool aosCmd::selectInputChannel(unsigned short _channel){
  channel=_channel;
  return true;
}

/* public method, assign input port (alternative to supplying channel freq) */
bool aosCmd::selectInputPort(unsigned short _port){
  iport=(unsigned short) pow(2L,_port-1);
  channel=iport2channel(iport);
  return true;
}


/* public method, return the currently selected input channel */
unsigned short aosCmd::getInputChannel(){
  return channel;
}

/* public method, return the currently selected input port */
unsigned short aosCmd::getInputPort(){
  return channel2iport();
}

/* public method, assign integration time */
bool aosCmd::setIntTime(unsigned short _intTime){
  intTime=_intTime;
  return true;
}

/* public method, assign laser current setting */
bool aosCmd::setLaserCurrent(float _laserCurrent){
  laserCurrent=_laserCurrent;
  return true;
}

/* public method, assign temperature control level */
bool aosCmd::setTempControl(float _tempControl){
  tempControl=_tempControl;
  return true;
}

/* public method, assign input attenuator level */
bool aosCmd::setAttenuator(unsigned short _attenuator){
  attenuator=_attenuator;
  return true;
}

/* private method, construct command */
char *aosCmd::cmd(unsigned short addr, unsigned short data, char *comment){
  strstream *mkComment = new strstream;
  *mkComment << "AOS:"
	     << comment
	     << ends;

  OdinCmd *ocom=new OdinCmd;
  return ocom->cmd(AOSUSR,addr,data,mkComment->str());
}

/* public method, construct permission command */
char *aosCmd::permitCmd(){
  return cmd(AOSPERMISSION,AOSPERMISSION_YES,"Permission to accept commands");
}

/* public method, construct no-permission command */
char *aosCmd::noPermitCmd(){
  return cmd(AOSPERMISSION,AOSPERMISSION_NO,"No command permission");
}

/* public method, standby command */
char *aosCmd::standbyCmd(){
  return cmd(AOSCMD,AOSCMD_STANDBY,"Standby mode");
}

/* public method, AOS comb calibration command */
char *aosCmd::combCmd(){
  return cmd(AOSCMD,AOSCMD_COMB,"Internal calibration");
}

/* public method, start integrating command */
char *aosCmd::goCmd(){
  return cmd(AOSCMD,AOSCMD_GO,"Go!");
}

/* public method, gain level selection command */
char *aosCmd::gainSelectCmd(){
  strstream *mkComment=new strstream;
  *mkComment << "Continuum value selected for gain search ("
	     << dec << setw(2) << setprecision(2)
	     << 10.0*log10(AOS_GAINSELECT_LEVEL/4095.0)
	     << "dB below saturation)"
	     << ends;
  return cmd(AOS_GAINSELECT,AOS_GAINSELECT_LEVEL,mkComment->str());
}

/* public method, gain search command */
char *aosCmd::gainSearchCmd(){
  return cmd(AOSCMD,AOSCMD_GAINSEARCH,"Attenuator gain search with CCD");
}

/* public method, laser current command */
char *aosCmd::laserCurrentCmd(){
  // until this is verified, use the fixed value 0x8C = 86.24 mA
  unsigned short cmdVal=
    (unsigned short) (AOS_LASER_OFFSET+4095.0*laserCurrent*AOS_LASER_SCALE);
  cmdVal = 0x8c;
  strstream *mkComment=new strstream;
  *mkComment << "Laser diode current: "
	     << dec << setw(5) << setprecision(5) << setfill('0')
	     << laserCurrent
	     << "mA"
	     << ends;
  return cmd(AOS_LASER,cmdVal,mkComment->str());
}

/* public method, construct No Input command (used for comb cal) */
char *aosCmd::noInputCmd(){
  return cmd(AOS_INPUT, AOS_NOINPUT,"No Input");
}

/* public method, construct input channel command */
char *aosCmd::inputCmd(){
  strstream *mkComment=new strstream;
  if (channel==NOINPUT)
    *mkComment << "No Input" << ends;
  else
    *mkComment << "Input Channel "
	       << dec << channel
	       << ends;
  return cmd(AOS_INPUT, channel2iport(), mkComment->str());
}

/* public method, construct synch mask command */
char *aosCmd::maskCmd(unsigned short whichMask, unsigned short mask){
  strstream *mkCommand=new strstream;

  if ((whichMask & MASK_ACDC1) == MASK_ACDC1)
    *mkCommand << cmd(AOS_MASKACDC1,mask,"ACDC 1 Mask");

  if ((whichMask & MASK_ACDC2) == MASK_ACDC2)
    *mkCommand << cmd(AOS_MASKACDC2,mask,"ACDC 2 Mask");

  if ((whichMask & MASK_CHOPA) == MASK_CHOPA)
    *mkCommand << cmd(AOS_MASKCHOPA,mask,"Chopper mech A Mask");

  if ((whichMask & MASK_CHOPB) == MASK_CHOPB)
    *mkCommand << cmd(AOS_MASKCHOPB,mask,"Chopper mech B Mask");

  if ((whichMask & MASK_119) == MASK_119)
    *mkCommand << cmd(AOS_MASK119,mask,"119 frequency switch Mask");

  *mkCommand << ends;
  return mkCommand->str();
}

/* public method, construct command to clear all masks */
char *aosCmd::clearMasksCmd(){
  return maskCmd(MASK_ACDC1 + MASK_ACDC2 + MASK_CHOPA + MASK_CHOPB + MASK_119,0);
}

/* private method, construct manual attenuator command */
char *aosCmd::attCmd(unsigned short dB){
  strstream mkComment;
  mkComment << "Attenuator level: "
	    << dec << setw(2) << setprecision(2) 
	    << dB
	    << ends;
  return cmd(AOS_ATT,dB,mkComment.str());
}

/* public method, return manual attenuator command */
char *aosCmd::attenuatorCmd(unsigned short dB){
  return attCmd(dB);
}

/* public overloaded method, return manual attenuator command
 * using preset attenuator level
 */
char *aosCmd::attenuatorCmd(){
  return attCmd(attenuator);
}

/* public method, construct AOP temperature control command */
char *aosCmd::temperatureControlCmd(){
  strstream *mkComment = new strstream;
  double cmdVal, valFrac;
  /* get the corresponding integer control level for a temperature, *
   * and round up                                                   */
  valFrac = modf(AOS_TEMPCTL_OFFSET+AOS_TEMPCTL_SCALE*tempControl,&cmdVal);
  if (valFrac >= 0.5) cmdVal+=1;

  /* re-evaluate the temperature control setting                    *
   * based on the integer command                                   */
  tempControl=(cmdVal-AOS_TEMPCTL_OFFSET)/AOS_TEMPCTL_SCALE;

  *mkComment << "AOP temperature control: "
	     << dec << setw(5) << setprecision(5) 
	     << tempControl
	     << " C" 
	     << ends;
  return cmd(AOS_TEMPCTL,(unsigned short) cmdVal,mkComment->str());
}

/* public method, construct integration time command */
char *aosCmd::intTimeCmd(){
  strstream *mkComment = new strstream;
  *mkComment << "Integration time: "
	     << dec << setw(2) << setprecision(2) 
	     << intTime
	     << ends;
  return cmd(AOS_INTTIME,intTime,mkComment->str());
}

/* public method, construct integration time command for use with comb cal*/
char *aosCmd::intTimeCombCmd(){
  strstream *mkComment = new strstream;
  *mkComment << "Integration time: "
	     << dec << setw(2) << setprecision(2) 
	     << 4
	     << ends;
  return cmd(AOS_INTTIME,4,mkComment->str());
}


/* public method, construct number of sub-integrations command */
char *aosCmd::subIntCmd(unsigned short n){
  strstream *mkComment = new strstream;
  *mkComment << "sub integrations: "
	     << dec << setw(2) << setprecision(2) 
	     << n
	     << ends;
  return cmd(AOS_SUBINT,n,mkComment->str());
}

/* public method, output mode command */
char *aosCmd::outputModeCmd(unsigned short mode){
  switch (mode) {

  case AOS_MODE_16:
  case AOS_MODE_16_ALSO:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_16,
	       "output mode 16 bit per pixel");
  break;

  case AOS_MODE_BIN2:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_BIN2,
	       "output mode 32-bit, binned over 2 pixels");
    break;

  case AOS_MODE_32:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_32,
	       "output mode 32-bit all pixels");
    break;

  case AOS_MODE_BIN4:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_BIN4,
	       "output mode binned over 4 pixels");
    break;

  case AOS_MODE_HIGH_RES_CENTRE:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_HIGH_RES_CENTRE,
	       "output mode high resolution in centre band");
    break;

  case AOS_MODE_HIGH_RES_WINGS:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_HIGH_RES_WINGS,
	       "output mode high resolution in the wings");
    break;

  case AOS_MODE_HIGH_RES_ADJUSTABLE:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_32,
	       "output mode high resolution in an adjustable band");
    break;

  default:
    return cmd(AOS_OUTPUTMODE,AOS_MODE_32,
	       "DEFAULT output mode 32-bit all pixels");
    break;
  }
  return "! error: no output mode assigned";
}

/* public method, produce AOS configuration commands
 * setting up the AOS, and preparing for a Gain Search
 */
char *aosCmd::configurationCmd(){
  strstream *mkSeq=new strstream;

  *mkSeq 
    << "\n"
    << "! AOS\n"
    << "! SET UP FOR A COMB CALIBRATION\n"
    << permitCmd()
    << noInputCmd()
    << attenuatorCmd(AOS_MAX_ATT)
    << laserCurrentCmd()
    /* AOS temperature control is set in the power-on script  *
     * to re-enable temperature control in the backend script *
     * uncomment to following line                            */
    /*    << temperatureControlCmd() */
    << intTimeCombCmd()
    << subIntCmd(1)
    << outputModeCmd(AOS_MODE_32)

    << "\n! MASK SETTINGS, NO SYNCHRO WHILE DOING COMB CAL\n"
    << clearMasksCmd()

    << "\n! EXECUTE COMB CAL\n"
    << combCmd()
    << "wait(9)\n"

    << "\n! RECONFIGURE AOS FOR CORRECT INPUT CHANNEL AND SETUP FOR CALIBRATION\n"
    << standbyCmd()
    << outputModeCmd(outputMode)
    << inputCmd()

    << ends;
  return mkSeq->str();
}

/* public method, AOS gain search commands */
char *aosCmd::calibrateCmd(){
  strstream *mkSeq=new strstream;
  *mkSeq
    << gainSelectCmd()
    << gainSearchCmd()
    << ends;
  return mkSeq->str();
}
