/* $Id: aosCmd.h for Odin
 * $auth: Steve Torchinsky
 * $created: Calgary-to-Toronto, Tuesday, 2001 May 08, 10:53:13
 *
 * This is the part of the cpp version of mkbckseq
 * mkbckseq is a Bash script employing sed and gawk which
 * uses templates to create backend command scripts
 *
 * The cpp version will do the same job, and will also read
 * the mode configurations from text files exported from the
 * Radiometer_modes.xls excel spreadsheet by Urban Frisk
 *
 * 
 * The class template for aosCmd is defined here.
 *
 * changes: 
 *  Calgary, Monday, 2001 September 24, 12:15:19
 *   AOS temperature control scale factors are corrected
 *
 *  Calgary, Wednesday, 2002 February 20, 11:04:44
 *   added macros definitions for attenuator settings: MIN,MAX,NOMINAL
 *
 *  Calgary, Wednesday, 2002 February 20, 16:02:26
 *   added method setAttenuator() and attenuator private variable
 *
 *  Calgary, Wednesday, 2002 February 20, 17:05:24
 *   new private method attCmd()
 *   method attenuatorCmd() is overloaded
 */

#include "../OdinTM.h"
#include "OdinCmd.h"


#ifndef AOSCMD_H
#define AOSCMD_H

/* AOS is user 6 */
#define AOSUSR 6
#define AOS 6

/* command addresses */
#define AOSPERMISSION 0
#define AOS_LASER 1
#define AOS_TEMPCTL 2
#define AOS_INPUT 3
#define AOS_ATT 4
#define AOS_COMB 5
#define AOSCMD 6
#define AOS_INTTIME 7
#define AOS_SUBINT 8
#define AOS_OUTPUTMODE 9
#define AOS_GAINSELECT 0xA
#define AOS_MASKACDC1 0xFA
#define AOS_MASKACDC2 0xFB
#define AOS_MASKCHOPA 0xFC
#define AOS_MASKCHOPB 0xFD
#define AOS_MASK119 0xFE

/* AOS execution commands all with address 6 */
#define AOSCMD_STANDBY 0
#define AOSCMD_GO 1
#define AOSCMD_COMB 2
#define AOSCMD_GAINSEARCH 4

/* AOS permission to accept commands, address 0 */
#define AOSPERMISSION_NO 0 
#define AOSPERMISSION_YES 1 

/* AOS comb control, address 5 */
#define AOSCOMB_OFF 0
#define AOSCOMB_ON 1

/* AOS continuum value level selected for gain search */
#define AOS_GAINSELECT_LEVEL 0x800

/* SCALING TO BE VERIFIED */
/* AOS laser diode current scaling for commands */
#define AOS_LASER_SCALE 1.623376
#define AOS_LASER_OFFSET 0

/* AOS temperature control scaling for commands */
#define AOS_TEMPCTL_SCALE 2.7533
#define AOS_TEMPCTL_OFFSET -2.6311

/* input ports */
#define AOS_NOINPUT 0
#define AOS_495 4
#define AOS_549 8
#define AOS_555 1
#define AOS_572 2
#define AOS_119 0x10

/* output modes */
#define AOS_MODE_16_ALSO 0
#define AOS_MODE_16 1
#define AOS_MODE_BIN2 2
#define AOS_MODE_32 3
#define AOS_MODE_BIN4 4
#define AOS_MODE_HIGH_RES_CENTRE 5
#define AOS_MODE_HIGH_RES_WINGS 6
#define AOS_MODE_HIGH_RES_ADJUSTABLE 7

/* attenuator settings */
#define AOS_MIN_ATT 0
#define AOS_MAX_ATT 0xF
#define AOS_NOM_ATT 6

/* class definition for AOS commanding */
class aosCmd {
 public:
  aosCmd();
  ~aosCmd();

  bool selectInputChannel(unsigned short);
  bool selectInputPort(unsigned short);
  bool selectOutputMode(unsigned short);
  unsigned short getInputChannel();
  unsigned short getInputPort();
  bool setIntTime(unsigned short);
  bool setLaserCurrent(float);
  bool setTempControl(float);
  bool setAttenuator(unsigned short);

  char *inputCmd();
  char *noInputCmd();
  char *permitCmd();
  char *noPermitCmd();
  char *standbyCmd();
  char *gainSelectCmd();
  char *gainSearchCmd();
  char *calibrateCmd();
  char *combCmd();
  char *laserCurrentCmd();
  char *maskCmd(unsigned short, unsigned short);
  char *clearMasksCmd();
  char *attenuatorCmd(unsigned short);
  char *attenuatorCmd();
  char *temperatureControlCmd();
  char *intTimeCmd();
  char *intTimeCombCmd();
  char *subIntCmd(unsigned short);
  char *outputModeCmd(unsigned short);
  char *configurationCmd();
  char *goCmd();

 private:
  bool defaultSettings();
  unsigned short channel2iport();
  unsigned short iport2channel(unsigned short);
  char *cmd(unsigned short, unsigned short, char *);
  char *attCmd(unsigned short);

  unsigned short outputMode;
  unsigned short iport;
  unsigned short channel;
  unsigned short intTime;
  unsigned short attenuator;

  float laserCurrent;
  float tempControl;

};


#endif
