/* $Id: bckCmd.cpp for Odin
 * $auth: Steve Torchinsky
 * $created: Calgary, Wednesday, 2001 May 16, 21:02:50
 *
 * This is the part of the cpp version of mkbckseq
 * mkbckseq is a Bash script employing sed and gawk which
 * uses templates to create backend command scripts
 *
 * The cpp version will do the same job, and will also read
 * the mode configurations from text files exported from the
 * Radiometer_modes.xls excel spreadsheet by Urban Frisk
 *
 * methods for class bckCmd are defined here
 *
 * changes:
 * Calgary, Thursday, 2001 June 14, 20:41:39
 *  added 119 Frequency Switching mask and assign method
 *
 * Calgary, Thursday, 2001 July 26, 08:42:51
 *  added method ocomFinalCmd()
 *  for extra commands for aero modes which are inserted from a text file
 *  (see mail from GF, Thu, 26 Jul 2001 11:06:14 +0200)
 *
 * Calgary, Friday, 2001 August 03, 09:45:08
 *  fixed bug:  ends to finish off footer, previously producing garbage
 *              in the DOS version
 *
 * Calgary, Thursday, 2001 August 16, 17:00:14
 *  fixed bug:  correlators attached to Signal Combiner did not undergo
 *              self-calibration!
 *
 * Calgary, Wednesday, 2001 September 05, 11:24:57 
 *  fixed bug: typo on line 577 meant that AC2 wasn't undergoing 
 *             self-calibration
 *
 * Calgary, Wednesday, 2001 November 21, 10:55:55
 *  changed DESCRIPTION, according to request from Stefan Lundin (email today)
 *
 * Calgary, Friday, 2001 December 07, 09:30:37
 *  changed DESCRIPTION, according to request from Gunnar today.
 *
 * Calgary, Thursday, 2002 January 31, 08:15:07
 *  v2.4.3
 *  removed correlator boot commands, according to request from Gunnar today.
 *
 * Calgary, Monday, 2002 February 04, 11:08:24
 *  v2.4.4
 *  self calibration section moved into its own method in preparation for 
 *  the new Radiometer_modes file format which will included self-cal=yes/no
 *
 * Calgary, Wednesday, 2002 February 20, 15:15:15
 *  renaming methods selfCalibrationCmd etc to inputLevelCmd
 *   input levels are adjusted either by self calibration 
 *   or by manual attenuator settings
 *
 * Calgary, Monday, 2002 February 25, 16:05:02
 *  methods for assigning self cal modified to also assign attenuator settings
 *
 * Bordeaux-to-Paris, Sunday, 2002 October 20, 19:09:05
 *  v2.6.0
 *  adding observingStrategy
 *
 * Paris-to-Montreal, Monday, 2002 October 21, 09:14:54
 *  changing chop mask in backends to always sync, regardless of obs strategy
 */

#include <cstdlib>
#include <cstring>
#include <strstream>
#include <iomanip>
#include <ctime>

#include "bckCmd.h"

/* constructor */
bckCmd::bckCmd(){
  aos=new aosCmd;
  ac1=new acCmd(AC1);
  ac2=new acCmd(AC2);

  /* default settings */
  enablePointing1();
  disablePointing2();
  assignDiscipline(ASTRO);
  assignObservingStrategy(DICKE_SWITCHING);
  disableFilterbank();
}

/* destructor */
bckCmd::~bckCmd(){
}

/* public method, set the mode name */
bool bckCmd::assignModeName(char *_modeName){
  modeName=_modeName;
  return true;
}

/* public method, set the mode variation (a,b,c,...) */
bool bckCmd::assignModeVariation(char _modeVariation){
  modeVariation=_modeVariation;
  return true;
}

/* public method, assign the mechanism controller */
bool bckCmd::assignMechController(char controller){
  mechController=controller;
  return true;
}

/* public method, assign the sequence version number */
bool bckCmd::assignSeqVersion(short _seqVersion){
  seqVersion=_seqVersion;
  return true;
}

/* public method, assign discipline: either astro or strat */
bool bckCmd::assignDiscipline(unsigned short _discipline){
  if (_discipline==STRAT)
    discipline=STRAT;
  else
    discipline=ASTRO;
  return true;
}

/* public method, assign observing strategy: either Dicke or Position */
bool bckCmd::assignObservingStrategy(unsigned short strategy){
  switch (strategy){
  case POSITION_SWITCHING:
    obsStrategy=POSITION_SWITCHING;
    break;
  case F119_SWITCHING:
    obsStrategy=F119_SWITCHING;
    break;
  default: /* DICKE_SWITCHING */
    obsStrategy=DICKE_SWITCHING;
  }
  return true;
}

/* public method, assign the integration time */
bool bckCmd::assignIntTime(unsigned short _intTime){
  intTime=_intTime;
  return true;
}

/* public method, assign the chop mask */
bool bckCmd::assignChopMask(unsigned short _chopMask){
  chopMask=_chopMask;
  return true;
}

/* public method, assign 119 Frequency Switching mask */
bool bckCmd::assignF119Mask(unsigned short _F119Mask){
  F119Mask=_F119Mask;
  return true;
}

/* public method, assign the mode description */
bool bckCmd::assignDescription(char *_description){

  //if (description) delete [] description;

  if (_description){
    description=_description;
    return true;
  }
  
  strstream *mkDesc=new strstream;
  *mkDesc << modeName << modeVariation << ends;
  description=mkDesc->str();

  return true;
}

/* public method, assign the final commands */
bool bckCmd::assignFinalCmd(char *_finalCmd){
  finalCmd=_finalCmd;
  return true;
}

/* private method, return the description */
char *bckCmd::modeDescription(){
  //if (description) return description;
  //assignDescription(NULL);
  return description;
}

/* public method, enable AOS */
bool bckCmd::enableAOS(){
  usingAOS=ON;
  return true;
}

/* public method, disable AOS */
bool bckCmd::disableAOS(){
  usingAOS=OFF;
  return true;
}

/* public method, enable AC1 */
bool bckCmd::enableAC1(){
  usingAC1=ON;
  return true;
}

/* public method, disable AC1 */
bool bckCmd::disableAC1(){
  usingAC1=OFF;
  return true;
}
/* public method, enable AC2 */
bool bckCmd::enableAC2(){
  usingAC2=ON;
  return true;
}

/* public method, disable AC2 */
bool bckCmd::disableAC2(){
  usingAC2=OFF;
  return true;
}

/* public method, enable Filterbank */
bool bckCmd::enableFilterbank(){
  ac2->enableFilterbank();
  return true;
}

/* public method, disable Filterbank */
bool bckCmd::disableFilterbank(){
  ac2->disableFilterbank();
  return true;
}

/* public method, enable pointing with ACDC1 */
bool bckCmd::enablePointing1(){
  usingPointing1=ON;
  return true;
}

/* public method, disable pointing with ACDC1 */
bool bckCmd::disablePointing1(){
  usingPointing1=OFF;
  return true;
}

/* public method, enable pointing with ACDC2 */
bool bckCmd::enablePointing2(){
  usingPointing2=ON;
  return true;
}

/* public method, disable pointing with ACDC2 */
bool bckCmd::disablePointing2(){
  usingPointing2=OFF;
  return true;
}

/* public method, assign self cal for AOS */
bool bckCmd::assignSelfCalAOS(unsigned short selfCal, unsigned short att){
#ifdef DEBUG
  cerr << "\nassignSelfCalAOS:  selfCal is " << selfCal
       << " att=" << att;
#endif
  if(selfCal==OFF){
    selfCalAOS=OFF;
    aos->setAttenuator(att);
    return true;
  }
  
  selfCalAOS=ON;
  return true;
}

/* public method, assign self cal for AC1 */
bool bckCmd::assignSelfCalAC1(unsigned short selfCal,
			      unsigned short att1,
			      unsigned short att2,
			      unsigned short att3,
			      unsigned short att4){
#ifdef DEBUG
  cerr << "\nassignSelfCalAC1:  selfCal is " << selfCal
       << " att1=" << att1 
       << " att2=" << att2
       << " att3=" << att3
       << " att4=" << att4;
#endif

  if(selfCal==OFF){
    selfCalAC1=OFF;
    ac1->setAtt(0,att1);
    ac1->setAtt(1,att2);
    ac1->setAtt(2,att3);
    ac1->setAtt(3,att4);
    return true;
  }
  
  selfCalAC1=ON;
  return true;
}

/* public method, assign self cal for AC2 */
bool bckCmd::assignSelfCalAC2(unsigned short selfCal,
			      unsigned short att1,
			      unsigned short att2,
			      unsigned short att3,
			      unsigned short att4){
#ifdef DEBUG
  cerr << "\nassignSelfCalAC2:  selfCal is " << selfCal
       << " att1=" << att1 
       << " att2=" << att2
       << " att3=" << att3
       << " att4=" << att4;
#endif
  if(selfCal==OFF){
    selfCalAC2=OFF;
    ac2->setAtt(0,att1);
    ac2->setAtt(1,att2);
    ac2->setAtt(2,att3);
    ac2->setAtt(3,att4);
    return true;
  }
  selfCalAC2=ON;
  return true;
}

/* private method, return the pointing mask to use */
unsigned short bckCmd::pointingMask(unsigned short discipline, unsigned short channel){
  if (discipline==STRAT) return POINTING_STRAT;
  if (channel==CH119) return POINTING_ASTRO_119;
  return POINTING_ASTRO_SUBMM;
}

/* public method, return procedure name (filename) */
char *bckCmd::procedureName(){
  strstream *mkName=new strstream;
  *mkName << modeName
	  << "bck."
	  << modeVariation
	  << ends;
  return mkName->str();
}

/* public method, return macro name */
char *bckCmd::macroName(){
  strstream *mkName=new strstream;
  *mkName << modeName
	  << modeVariation
	  << "BCK"
	  << ends;
  return mkName->str();
}

/* public method, command to position chopper using OtS */
char *bckCmd::positionChopperCmd(unsigned short pos){
  strstream *mkComment = new strstream;
  *mkComment << "Ots:command chopper to ";
  if (pos==POSITION_A)
    *mkComment << "A-side on source";
  else {
    *mkComment << "B-side on source";
    pos=POSITION_B;
  }
  *mkComment << ends;

  unsigned short addr;
  if (mechController=='B')
    addr=OTS_MECHB;
  else
    addr=OTS_MECHA;

  OdinCmd *ocom=new OdinCmd;
  return ocom->cmd(OTS_USR,addr,pos,mkComment->str());
}

/* private method, command chopper to stop */
char *bckCmd::stopChopperCmd(){
  strstream *mkComment=new strstream;
  *mkComment << "Mech" << mechController << ":Stop Chopper" << ends;

  unsigned short usr;
  if (mechController=='B')
    usr=MECHB_USR;
  else
    usr=MECHA_USR;

  OdinCmd *ocom=new OdinCmd;
  return ocom->cmd(usr,MECH_CHOP,MECH_CHOP_STOP,mkComment->str());
}

/* private method, command to run the chopper */
char *bckCmd::runChopperCmd(){

  unsigned short usr;
  if (mechController=='B')
    usr=MECHB_USR;
  else
    usr=MECHA_USR;

  unsigned short data;
  data=MECH_CHOP_RUN | intTime;

  strstream *mkComment=new strstream;
  *mkComment << "Mech" << mechController 
	     << ":chopper low power forwards;step: "
	     << intTime
	     << "sec"
	     << ends;

  OdinCmd *ocom=new OdinCmd;
  return ocom->cmd(usr,MECH_CHOP,data,mkComment->str());
}


/* private method, the header of the sequence file
 * this is the same for all modes, regardless of which
 * backends are running
 */
char *bckCmd::ocomHeader(){
  strstream *mkSeq=new strstream;
  time_t *now=new time_t;
  time(now);
  char *seqDate=new char[80];
  strftime(seqDate,80,"%A, %Y %B %d, %X",gmtime(now));
  *mkSeq
    << "PROCEDURE   : \"" << procedureName() << "\"\n"

    << "DESCRIPTION : \"" << modeName << modeVariation
    << " Backend, Mech-" << mechController;
  if (usingPointing1 == ON)
    *mkSeq << " ACDC1";
  else if(usingPointing2 == ON)
    *mkSeq << " ACDC2";
  else
    *mkSeq << " NoACDC";
  
  *mkSeq
    << dec
    << " v" << MKBCKSEQ_VERSION_MAJOR 
    << "."  << MKBCKSEQ_VERSION_MINOR
    << "."  << MKBCKSEQ_VERSION_TINY
    << " modes" << seqVersion
    << "\"\n"

    << "VERSION     : \"1." << dec << seqVersion << "\"\n"

    << "AUTHOR      : \"Steve's mkbckseq generator v"
    << dec << MKBCKSEQ_VERSION_MAJOR 
    << "." << MKBCKSEQ_VERSION_MINOR
    << "." << MKBCKSEQ_VERSION_TINY
    << "\"\n"

    << "COMMENT     : \"" << seqDate << " UTC\"\n"

    << "! MACRONAME : " << macroName() << "\n"
    //<< "! MODE DESCRIPTION: " << modeDescription() << "\n"
    << "message = self\n"
    << "run\n"
    << "on invalid oper\n"
    << "delay=0\n\n"

    << "! log the satellite time word when this sequence begins\n"
    << "check(OSC02)\n"
    << "check(OSC01)\n"
    << "begin:\n\n"

    << "! THIS SEQUENCE ASSUMES THAT \n"
    << "! RADIOMETER UNITS ARE POWERED AND INITIALISED\n"
    << "! AND THAT THE FRONT-END IS UP AND RUNNING\n"
    << "!\n"
    << "! halt all backends before stopping the chopper\n"


    /* Calgary, Thursday, 2002 January 31, 08:15:07            */
    /* boot command for correlators no longer implemented here */
    /* << ac1->bootCmd()
     * << ac2->bootCmd()
     * << "wait(3)\n"
     */

    << ac1->haltCmd()
    << ac2->haltCmd()
    << "wait(11)\n"
    << ac1->configCmd()
    << ac2->configCmd()
    << "wait(1)\n"
    << aos->permitCmd()
    << aos->standbyCmd()
    << "! Laser Diode is NOT switched off in order to maintain its temperature\n"
    << "! see notes from ALX 2000 Dec 4\n"
    << aos->noInputCmd()
    << aos->attenuatorCmd(0xF)
    << aos->noPermitCmd()

    << "!\n!\n"

    << "! stop the chopper in order to do the input level calibration\n"
    << stopChopperCmd()
    << "!\n"
    << "! wait for chopper to finish chop\n"
    << "wait(10)\n"

    << ends;

  return mkSeq->str();
}

/* private method, AOS start normal observation commands */
char *bckCmd::aosStartCmd(){
  strstream *mkSeq=new strstream;
  *mkSeq << aos->standbyCmd();

  if (usingPointing1 == ON)
    *mkSeq << aos->maskCmd(MASK_ACDC1,
			   pointingMask(discipline,aos->getInputChannel()));
  if (usingPointing2 == ON)
    *mkSeq << aos->maskCmd(MASK_ACDC2,
			   pointingMask(discipline,aos->getInputChannel()));

  switch (obsStrategy){
  case DICKE_SWITCHING:
    aos->setIntTime(0);
    break;
  case F119_SWITCHING:
    aos->setIntTime(0);
    *mkSeq << aos->maskCmd(MASK_119,F119Mask);
    break;
  default:
    aos->setIntTime(intTime);
  }

  /* sync to chopper regardless of observing strategy */
  if (mechController=='B')
    *mkSeq << aos->maskCmd(MASK_CHOPB,chopMask);
  else
    *mkSeq << aos->maskCmd(MASK_CHOPA,chopMask);

  *mkSeq << aos->intTimeCmd()
	 << aos->subIntCmd(1)
	 << aos->goCmd()
	 << aos->noPermitCmd()
	 << ends;

  return mkSeq->str();
}

/* private method, Correlator start normal observation commands */
char *bckCmd::acStartCmd(unsigned short whichAC){
  strstream *mkSeq=new strstream;
  acCmd *ac;
  if (whichAC==AC2)
    ac=ac2;
  else
    ac=ac1;

  /* assign masks */
  if (usingPointing1 == ON)
    ac->setMask(MASK_ACDC1,pointingMask(discipline,ac->getInputChannel()));

  if (usingPointing2 == ON)
    ac->setMask(MASK_ACDC2,pointingMask(discipline,ac->getInputChannel()));

  if ((usingPointing1 == ON) ||  (usingPointing2 == ON))
    *mkSeq << ac->maskCmd(MASK_ACDC1);

  if (mechController=='B')
    ac->setMask(MASK_CHOPB,chopMask);
  else
    ac->setMask(MASK_CHOPA,chopMask);

  ac->setMask(MASK_119,F119Mask);

  switch (obsStrategy){
  case DICKE_SWITCHING:
    ac->setIntTime(0);
    ac->enableSwitching();
    *mkSeq << ac->maskCmd(MASK_CHOPA);
    break;
  case F119_SWITCHING:
    ac->setIntTime(0);
    ac->enableSwitching();
    *mkSeq << ac->maskCmd(MASK_119);
    break;
  default: /* POSITION_SWITCHING */
    ac->disableSwitching();
    ac->setIntTime(intTime);
    *mkSeq << ac->maskCmd(MASK_CHOPA);
  }

  *mkSeq << ac->intTimeCmd()
	 << ac->inputCmd()
	 << ac->configCmd()
	 << "wait(1)\n"
	 << ac->startCmd()
	 << ends;

  return mkSeq->str();
}

/* private method, AC1 start normal observation commands */
char *bckCmd::ac1StartCmd(){
  return acStartCmd(AC1);
}

/* private method, AC2 start normal observation commands */
char *bckCmd::ac2StartCmd(){
  return acStartCmd(AC2);
}

/* private method, backend input level settings for upper side */
char *bckCmd::inputLevelUpperCmd(){

  strstream mkSeq;
  strstream mkCalSeq;
  short wait=0;
  bool manualInputLevel=false;

  mkSeq 
    << "!------ Start: BACKEND INPUT CALIBRATION LEVEL UPPER SIDE------- !\n";

  if ((usingAC1==ON)
      && ((ac1->getInputChannel() == CH495)
	  || (ac1->getInputChannel() == CH549)
	  || (ac1->getInputChannel() == CHCOMBINER))){
    if (selfCalAC1==ON){
      mkCalSeq << "! CALIBRATE AC1\n"
	       << ac1->calibrateCmd();
      wait=10;
    }
    else {
      mkCalSeq << "! SET INPUT ATTENUATORS FOR AC1\n"
	       << ac1->attenuatorsCmd();
      manualInputLevel=true;
    }
  }

  if ((usingAC2==ON)
      && ((ac2->getInputChannel() == CH495)
	  || (ac2->getInputChannel() == CH549))){
    if (selfCalAC2==ON){
      mkCalSeq << "! CALIBRATE AC2\n"
	       << ac2->calibrateCmd();
      wait=10;
    }
    else {
      mkCalSeq << "! SET INPUT ATTENUATORS FOR AC2\n"
	       << ac2->attenuatorsCmd();
      manualInputLevel=true;
    }
  }

  if ((usingAOS==ON)
      && ((aos->getInputChannel() == CH495)
	  || (aos->getInputChannel() == CH549))){
    if (selfCalAOS==ON){
      mkCalSeq << "! CALIBRATE AOS\n"
	       << aos->calibrateCmd();
      wait=16;
    }
    else {
      mkCalSeq << "! SET INPUT ATTENUATOR FOR AOS\n"
	       << aos->attenuatorCmd();
      manualInputLevel=true;
    }
  }
  mkCalSeq << ends;

  if (wait>0){
    mkSeq << positionChopperCmd(POSITION_A)
	  << "wait(2) ! wait for Chopper to reach position\n"
	  << mkCalSeq.str()
	  << "wait("
	  << dec << setw(2) << setprecision(2)
	  << wait
	  << ") ! wait for calibration\n";
  }
  else if (manualInputLevel)
    mkSeq << mkCalSeq.str();

  mkSeq
    << "!------ End:   BACKEND INPUT CALIBRATION LEVEL UPPER SIDE------- !\n"
    << ends;
  return mkSeq.str();
}

/* private method, backend self calbration for lower side */
char *bckCmd::inputLevelLowerCmd(){
  strstream mkCalSeq;
  strstream mkSeq;
  short wait=0;
  bool manualInputLevel=false;

  mkSeq 
    << "!------ Start: BACKEND INPUT CALIBRATION LEVEL LOWER SIDE------- !\n";

  if ((usingAC1==ON)
      && ((ac1->getInputChannel() == CH555)
	  || (ac1->getInputChannel() == CH572)
	  || (ac1->getInputChannel() == CH119))){
    if (selfCalAC1==ON){
      mkCalSeq << "! CALIBRATE AC1\n"
	       << ac1->calibrateCmd();
      wait=10;
    }
    else {
      mkCalSeq << "! SET INPUT ATTENUATORS FOR AC1\n"
	       << ac1->attenuatorsCmd();
      manualInputLevel=true;
    }
  }

  if ((usingAC2==ON)
      && ((ac2->getInputChannel() == CH555)
	  || (ac2->getInputChannel() == CH572)
	  || (ac2->getInputChannel() == CH119)
	  || (ac2->getInputChannel() == CHCOMBINER))){
    if (selfCalAC2==ON){
      mkCalSeq << "! CALIBRATE AC2\n"
	       << ac2->calibrateCmd();
      wait=10;
    }
    else {
      mkCalSeq << "! SET INPUT ATTENUATORS FOR AC2\n"
	       << ac2->attenuatorsCmd();
      manualInputLevel=true;
    }
  }

  if ((usingAOS==ON)
      && ((aos->getInputChannel() == CH555)
	  || (aos->getInputChannel() == CH572)
	  || (aos->getInputChannel() == CH119))){
    if (selfCalAOS==ON){
      mkCalSeq << "! CALIBRATE AOS\n"
	       << aos->calibrateCmd();
      wait=16;
    }
    else {
      mkCalSeq << "! SET INPUT ATTENUATOR FOR AOS\n"
	       << aos->attenuatorCmd();
      manualInputLevel=true;
    }
  }
  mkCalSeq << ends;

  if (wait>0){
    mkSeq << positionChopperCmd(POSITION_B)
	   << "wait(2) ! wait for Chopper to reach position\n"
	   << mkCalSeq.str()
	   << "wait("
	   << dec << setw(2) << setprecision(2)
	   << wait
	   << ") ! for calibration\n";
  }
  else if (manualInputLevel)
    mkSeq << mkCalSeq.str();
  

  mkSeq
    << "!------ End:   BACKEND INPUT CALIBRATION LEVEL LOWER SIDE------- !\n"
    << ends;

  return mkSeq.str();
}

/* private method, backend input level setting of input attenuators */
char *bckCmd::inputLevelCmd(){
  strstream mkSeq;
  mkSeq
    << "\n\n!---------------------------------------------------\n"
    << "! BACKEND INPUT LEVEL CALIBRATION\n"
    << "!\n! Once all backends have been configured\n"
    << "! we can calibrate their input levels.\n"
    << "! In order to save wait states,\n"
    << "! do the backends together that are looking\n"
    << "! at the same front-end beam path\n\n"

    << inputLevelUpperCmd()
    << endl
    << inputLevelLowerCmd()
    << ends;

  return mkSeq.str();
}

/* private method, ocom seqence foot */
char *bckCmd::ocomFooter(){
  strstream *mkSeq = new strstream;
  *mkSeq
    << finalCmd
    << "\n\n\nend:\n"
    << "! log the satellite time word when this sequence ends\n"
    << "check(OSC02)\n"
    << "check(OSC01)\n"
    << "\nexit\n"
    << "EOF\n"
    << ends;

 return mkSeq->str();
}


/* public method, the complete backend sequence */
char *bckCmd::ocomSequence(){
  strstream *mkSeq=new strstream;

  *mkSeq << ocomHeader();

  if (usingAC1 == ON)
    *mkSeq << ac1->configurationCmd();
  else
    *mkSeq << "!AC1 is not in use in this mode\n";

  if (usingAC2 == ON)
    *mkSeq << ac2->configurationCmd();
  else
    *mkSeq << "!AC2 is not in use in this mode\n";

  if (usingAOS == ON)
    *mkSeq << aos->configurationCmd();
  else
    *mkSeq << "!AOS is not in use in this mode\n";

  /* input level setting of backend input attenuators
   * it's either done manually or by self calibration 
   */
  *mkSeq << inputLevelCmd();

  if (obsStrategy==DICKE_SWITCHING)
    *mkSeq << "\n\n"
	   << "\n! CHOPPER CONFIGURATION DURING MODE\n"
	   << runChopperCmd()
	   << "wait(1)";

  *mkSeq << "\n! START THE BACKENDS\n";

  if (usingAC1 == ON) *mkSeq << ac1StartCmd();
  *mkSeq << endl;
  if (usingAC2 == ON) *mkSeq << ac2StartCmd();
  *mkSeq << endl;
  if (usingAOS == ON) *mkSeq << aosStartCmd();
  *mkSeq << endl;

  *mkSeq << ocomFooter()
	 << ends;

  return mkSeq->str();
}
