/* $Id: mkBckSeq.cpp for Odin
 * $auth: Steve Torchinsky
 * $created: Calgary, Friday, 2001 June 01, 11:58:37
 *
 * This is the cpp version of mkbckseq
 * mkbckseq is a Bash script employing sed and gawk which
 * uses templates to create backend command scripts
 *
 * The cpp version will do the same job, and will also read
 * the mode configurations from text files exported from the
 * Radiometer_modes.xls excel spreadsheet by Urban Frisk
 *
 * method definitions for class mkBckSeq
 * 
 * changes:  Calgary, Monday, 2001 June 11, 07:14:44
 *           integration time 4 secs for Aero and 10 secs for Astro
 *           except a12 which is 5 secs. 
 *           based on email from Mathias today.
 *
 *           Calgary, Monday, 2001 September 24, 12:17:38
 *           AOS temperature control level selection 
 *           is now in method assignMode, but still not read as a parameter
 *           from the Radiometer Modes spreadsheet.
 *
 *           Calgary, Friday, 2001 November 09, 08:46:10
 *           AOS temperature control level 22C
 *
 *           Calgary, Thursday, 2001 December 06, 16:24:14
 *           invalid modes are skipped, and further sequences are generated
 *           instead of immediately exiting the program
 *
 *           Bordeaux-to-Paris, Sunday, 2002 October 20, 19:15:35
 *           v2.6.0
 *           adding Observing Strategy
 *
 *           Paris-to-Montreal, Monday, 2002 October 21, 09:24:12
 *           always sync to chopper, regardless of observing strategy
 */


#include <cstdlib>
#include <iostream>
#include <strstream>
#include <iomanip>
#include <cstring>

#include "mkBckSeq.h"


/* constructor */
mkBckSeq::mkBckSeq(){
  modes=new smrModes;
}

/* destructor */
mkBckSeq::~mkBckSeq(){
}

/* public method, read the Radiometer Modes */
bool mkBckSeq::readModes(){
  return modes->readModes();
}

/* private method, assign the mode parameters */
bool mkBckSeq::assignMode(short n){
  bck=new bckCmd;

  /* assign mode name */
  if(!bck->assignModeName(modes->modeName(n)))return false;


  /* assign mode variation */
  if(!bck->assignModeVariation(modes->modeVariation(n)))return false;


  /* assign integration time (usually chopper time) */
  /* Calgary, Tuesday, 2001 June 05, 14:28:16
   * this does not appear on the Radiometer Modes file
   * so we hard wire it here for now 
   * (I know that comets a12 is 5 seconds and everything else
   * is 10 seconds)
   *
   * Calgary, Monday, 2001 June 11, 07:14:44
   *  Aero is 4 seconds
   *
   * Calgary, Thursday, 2001 June 14, 20:18:46
   *  these are now assigned in class smrModes
   *  they are read in the file, or else the default settings are used
   */
  if(!bck->assignIntTime(modes->intTime(n)))return false;

  /* assign observing strategy */
  if(!bck->assignObservingStrategy(modes->ObservingStrategy(n)))return false;

  /* assign mech controller */
  if (mechController==MECHB){
    if(!bck->assignMechController('B'))return false;
  } 
  else 
    if(!bck->assignMechController('A'))return false;

  /* assign acdc controller, if any */
  switch (acdc) {
  case ACDC1: 
    if(!bck->enablePointing1())return false;
    if(!bck->disablePointing2())return false;
    break;
  case ACDC2: 
    if(!bck->enablePointing2())return false;
    if(!bck->disablePointing1())return false;
    break;
  default:
    if(!bck->disablePointing1())return false;
    if(!bck->disablePointing2())return false;
  }

  /* assign sequence version number
   * we use the version number of the Radiometer Modes file 
   */
  if(!bck->assignSeqVersion(modes->versionNumber()))return false;

  /* assign discipline */
  if(!bck->assignDiscipline(modes->Discipline(n)))return false;

  /* assign chop mask: always sync to chopper */
  if(!bck->assignChopMask(CHOP_SYNC))return false;
  /**********************************************************
  if (modes->chopSync(n)==ON){
    if(!bck->assignChopMask(CHOP_SYNC))return false;
  }
  else
    if(!bck->assignChopMask(0))return false;
  **********************************************************/

  /* assign 119 Frequency Switch mask */
  if (modes->F119Sync(n)==ON){
    if(!bck->assignF119Mask(F119_SYNC))return false;
  }
  else
    if(!bck->assignF119Mask(0))return false;

  /* assign mode description */
  /* this causes problems with the DOS version */
  if(!bck->assignDescription(modes->modeDescription(n)))return false;


  /* enable/disable backends and assign input and mode of operation */

  /* AOS */
  if (modes->usingAOS(n)==OFF){
    if(!bck->disableAOS())return false;
  }
  else {
    if(!bck->enableAOS())return false;
    if(!bck->aos->selectInputPort(modes->aosInput(n)))return false;
    if(!bck->aos->selectOutputMode(modes->aosMode(n)))return false;
    if(!bck->aos->setTempControl(22.0))return false;
    if(!bck->assignSelfCalAOS(modes->aosSelfCal(n),
			      modes->aosATT(n)))return false;
  }


  /* AC1 */
  if (modes->usingAC1(n)==OFF){
    if(!bck->disableAC1())return false;
  }
  else {
    if(modes->fileFormat()>=2) 
      if(!bck->ac1->setMode(modes->ac1Mode(n),
			    modes->ac1Mux(n),
			    modes->ac1SSB1(n),
			    modes->ac1SSB2(n),
			    modes->ac1SSB3(n),
			    modes->ac1SSB4(n),
			    modes->modeDescription(n)))return false;
    if(!bck->enableAC1())return false;
    if(!bck->ac1->selectInputPort(modes->ac1Input(n)))return false;
    if(!bck->ac1->selectMode(modes->ac1Mode(n)))return false;
#ifdef DEBUG
    cerr << "\nassignMode  ac1SelfCal(n)="
	 << modes->ac1SelfCal(n);
#endif
    if(!bck->assignSelfCalAC1(modes->ac1SelfCal(n),
			      modes->ac1ATT1(n),
			      modes->ac1ATT2(n),
			      modes->ac1ATT3(n),
			      modes->ac1ATT4(n)))return false;
  }


  /* AC2 */
  if (modes->usingAC2(n)==OFF){
    if(!bck->disableAC2())return false;
  }
  else {
    if(modes->fileFormat()>=2)
      if(!bck->ac2->setMode(modes->ac2Mode(n),
			    modes->ac2Mux(n),
			    modes->ac2SSB1(n),
			    modes->ac2SSB2(n),
			    modes->ac2SSB3(n),
			    modes->ac2SSB4(n),
			    modes->modeDescription(n)))return false;
    if(!bck->enableAC2())return false;
    if(!bck->ac2->selectInputPort(modes->ac2Input(n)))return false;
    if(!bck->ac2->selectMode(modes->ac2Mode(n)))return false;
    if(!bck->assignSelfCalAC2(modes->ac2SelfCal(n),
			      modes->ac2ATT1(n),
			      modes->ac2ATT2(n),
			      modes->ac2ATT3(n),
			      modes->ac2ATT4(n)))return false;
  }


  /* Filterbank */
  if (modes->usingFilterbank(n)==OFF){
    if(!bck->disableFilterbank())return false;
  } 
  else
    if(!bck->enableFilterbank())return false;

  /* assign final commands */
  if(!bck->assignFinalCmd(modes->finalCmd(n)))return false;


  return true;
}

/* private method, check for a directory existence
 * and create it if it doesn't exist
 */
bool mkBckSeq::mkdir(char *dirname){
  strstream *fname=new strstream;
  *fname << dirname 
	 << SLASH << "mkbckseq.tmp"
	 << ends;

  /* return if temporary file successfully created */
  ofstream *tmpfile=new ofstream;
  tmpfile->open(fname->str(),ios::trunc | ios::out);
  delete fname;
  if (!tmpfile->fail()){
    tmpfile->close();
    return true;
  }

  /* try to make the directory */
  strstream *cmd=new strstream;
  *cmd << "mkdir "
	 << dirname
	 << ends;
  int ret=system(cmd->str());
  if (ret==0) return true;
  else return false;
}

/* private method, open the sequence file for output
 * the acdc and mech controller subdirectories must exist or be created */
bool mkBckSeq::openSeqFile(){

  char *acdcDirName;
  char *mechCtlDirName;

  switch (acdc) {
  case ACDC1:  acdcDirName="acdc1";
    break;
  case ACDC2: acdcDirName="acdc2";
    break;
  default: acdcDirName="noacdc";
  }

  switch (mechController){
  case MECHA: mechCtlDirName="A";
    break;
  case MECHB: mechCtlDirName="B";
    break;
  default: mechCtlDirName="error";
  }

  if (!mkdir(acdcDirName)){
    cout << "\ncould not create directory: "
	 << acdcDirName
	 << endl;
    return false;
  }

  strstream *dirname=new strstream;
  *dirname << acdcDirName
	   << SLASH
	   << mechCtlDirName
	   << ends;

  if (!mkdir(dirname->str())){
    cout << "\ncould not create directory: "
	 << dirname->str()
	 << endl;
    return false;
  }
  delete dirname;

  /* directory structure is okay, so we can create the output file */
  strstream *fname=new strstream;
  *fname << acdcDirName
	 << SLASH
	 << mechCtlDirName
	 << SLASH
	 << bck->procedureName()
	 << ends;

  bckseq=new ofstream;
  bckseq->open(fname->str(),ios::trunc | ios::out);
  if (bckseq->fail()){
    cout << "\ncould not create file: "
	 << fname->str()
	 << endl;
    return false;
  }


  return true;
}

/* private method, make all the backend sequence files 
 * for a given mech controller and acdc controller
 */
bool mkBckSeq::mkSequences(){
  short n;

  for (n=0;n<modes->nModes();n++){
    if (!assignMode(n)){
      cout << "\n        "
	   << modes->modeName(n)
	   << modes->modeVariation(n)
	   << " mode not valid";
      continue;
    }

    if (!openSeqFile())return false;
    *bckseq << bck->ocomSequence();
    bckseq->close();

    delete bckseq;
    delete bck;
  }

  return true;
}

/* public method, make all the sequences for all the modes
 * and for both mech and acdc controllers
 */
bool mkBckSeq::mkAllSequences(){
  if (!readModes())return false;

  cout << "\nmaking sequences with ACDC1 and MechA ..." << flush;
  acdc=ACDC1;
  mechController=MECHA;
  if (!mkSequences())return false;
  cout << " done" << flush;

  cout << "\nmaking sequences with ACDC1 and MechB ...";
  acdc=ACDC1;
  mechController=MECHB;
  if (!mkSequences())return false;
  cout << " done";

  cout << "\nmaking sequences with ACDC2 MechA ...";
  acdc=ACDC2;
  mechController=MECHA;
  if (!mkSequences())return false;
  cout << " done";

  cout << "\nmaking sequences with ACDC2 MechB ...";
  acdc=ACDC2;
  mechController=MECHB;
  if (!mkSequences())return false;
  cout << " done";

  cout << "\nmaking sequences with no ACDC and with MechA ...";
  acdc=NOACDC;
  mechController=MECHA;
  if (!mkSequences())return false;
  cout << " done";

  cout << "\nmaking sequences with no ACDC and with MechB ...";
  acdc=NOACDC;
  mechController=MECHB;
  if (!mkSequences())return false;
  cout << " done";

  cout << endl;
  return true;
}

